/*
 * Adding two arrays in parallel with C++ threads.
 * 
 * This demo illustrates the following:
 * - data parallel computation
 * - distributing a non-overlapping range of inputs/outputs to each thread 
 * - speedup from parallelism (run with 1,2,3... threads)
 *
 * CSE 374, 2015-03 BM
 */

#include <thread>
#include <iostream>
#include <string>

using namespace std;

// represents the integer range [start, start+len)
struct range {
  int start;
  int len;
};

// do out <-- in1 + in2 for only the specified range of the arrays
void add_region(int* in1, int* in2, int* out, range r) {
  for (int i=r.start; i<r.len; i++) {
    out[i] = in1[i] + in2[i];
  }
}

// Calculate a 1-dimensional block distribution of n elements into b blocks 
// n number of elements
// b number of blocks
// id the block number asked for
// Assumes b <= n
range block_1d(int n, int b, int id) {
  int base_per_proc = n / b;
  int remainder = n % b;

  range result;

  // every block gets at least base_per_proc elements 

  // give +1 to each block where id < remainder
  if (id < remainder) {
    result.start = (base_per_proc+1)*id;
    result.len = base_per_proc + 1;
  } else {
    result.start = (base_per_proc+1)*remainder + base_per_proc*(id-remainder);
    result.len = base_per_proc;
  }

  return result;
}


int main(int argc, char** argv) {
  // usage: ./parallel_array_add [number of threads]
  int p = 1;
  if (argc >= 2) { 
    p = stoi(argv[1]);
  }

  // length of the arrays
  const int N = 100000000;

  // we will calculate z = x + y;
  // So allocate the input and output arrays
  int* x = new int[N];
  int* y = new int[N];
  int* z = new int[N];

  /* assume x and y are initialized to something interesting... */

  // start p threads
  thread** ts = new thread*[p];
  for (int i = 0; i<p; i++) {
    // assign a range of the arrays to thread i
    range r = block_1d(N, p, i);
    // create and start thread i
    ts[i] = new thread(add_region, x, y, z, r);
  }

  // wait for all threads to complete
  for (int i = 0; i<p; i++) {
    // wait for thread i to complete
    ts[i]->join();

    // done with thread i, so delete it
    delete ts[i];
  }

  delete[] ts;
}
